import math
import numpy as np
try:
  import cupy
  import optimized_gpu_tools
  import Q_inner_product
  import SphericalHarmonics_gpu
  xpy_default=cupy
except:
  print(' no cupy')
  import numpy as cupy
  optimized_gpu_tools=None
  Q_inner_product=None
  xpy_default=np


_coeff2m2 = math.sqrt(5.0 / (64.0 * math.pi))
_coeff2m1 = math.sqrt(5.0 / (16.0 * math.pi))
_coeff20 = math.sqrt(15.0 / (32.0 * math.pi))
_coeff2p1 = math.sqrt(5.0 / (16.0 * math.pi))
_coeff2p2 = math.sqrt(5.0 / (64.0 * math.pi))

def SphericalHarmonicsVectorized_orig(
        lm,
        theta, phi,
        xpy=cupy, dtype=np.complex128,
    ):
    """
    Compute spherical harmonics Y_{lm}(theta, phi) for a given set of lm's,
    thetas, and phis.

    Parameters
    ----------
    lm : array_like, shape = (n_indices, 2)

    theta : array_like, shape = (n_params,)

    phi : array_like, shape = (n_params,)

    Returns
    -------
    Ylm : array_like, shape = (n_params, n_indices)
    """
    l, m = lm.T

    n_indices = l.size
    n_params = theta.size

    Ylm = xpy.empty((n_params, n_indices), dtype=dtype)

    # Ensure coefficients are on the board if using cupy.
    c2m2 = xpy.asarray(_coeff2m2)
    c2m1 = xpy.asarray(_coeff2m1)
    c20 = xpy.asarray(_coeff20)
    c2p1 = xpy.asarray(_coeff2p1)
    c2p2 = xpy.asarray(_coeff2p2)
    imag = xpy.asarray(1.0j)

    # Precompute 1 +/- cos(theta).
    cos_theta = xpy.cos(theta)
    one_minus_cos_theta = (1.0 - cos_theta)
    one_plus_cos_theta = xpy.add(1.0, cos_theta, out=cos_theta)
    del cos_theta

    # Precompute sin(theta).
    sin_theta = xpy.sin(theta)

    for i, m_i in enumerate(m):
        if m_i == -2:
            Ylm[...,i] = (
                c2m2 *
                xpy.square(one_minus_cos_theta) *
                xpy.exp(imag * m_i * phi)
            )
        elif m_i == -1:
            Ylm[...,i] = (
                c2m1 *
                xpy.multiply(sin_theta, one_minus_cos_theta) *
                xpy.exp(imag * m_i * phi)
            )
        elif m_i == 0:
            Ylm[...,i] = (
                c20 *
                xpy.square(sin_theta)
            )
        elif m_i == +1:
            Ylm[...,i] = (
                c2p1 *
                xpy.multiply(sin_theta, one_plus_cos_theta) *
                xpy.exp(imag * m_i * phi)
            )
        elif m_i == +2:
            Ylm[...,i] = (
                c2p2 *
                xpy.square(one_plus_cos_theta) *
                xpy.exp(imag * m_i * phi)
            )
        else:
            Ylm[...,i] = xpy.nan

    return Ylm



#! /usr/bin/env python
#  
#
# Tool: dumps a candidate waveform to a frame file.
# Default GPS time used is boring
# Tries to be as close to C as possible -- no interface via pylal/glue
#
# EXAMPLE
#    python util_LALWriteFrame.py; FrDump -i dummy.gwf
#
#  WARNING: My version does NOT interpolate the signal to a synchronized set of sample times.
#                       This may cause problems for some applications, particularly at low sample rates.


import argparse
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal

parser = argparse.ArgumentParser()
parser.add_argument("--fname", default=None, help = "Base name for output frame file. Otherwise auto-generated ")
parser.add_argument("--background-cache",default=None,help="If nonzero, loads a frame cache, and overlays with result. Used to add noise. Must specify background-channel ")
parser.add_argument("--background-channel",default=None,help="If nonzero, loads a frame cache, and overlays with result. Used to add noise. Must specify background-cache ")
parser.add_argument("--background-add-from-gaussian-psd",action="store_true",help="Adds recolored gaussian noise, based on a specific PSD file")
parser.add_argument("--psd-file",action='append')
parser.add_argument("--instrument", default="H1",help="Use H1, L1,V1")
parser.add_argument("--inj", dest='inj', default=None,help="inspiral XML file containing injection information.")
parser.add_argument("--event",type=int, dest="event_id", default=None,help="event ID of injection XML to use.")
parser.add_argument("--single-ifo",default=False,action='store_true')
parser.add_argument("--approx",type=str,default=None)
parser.add_argument("--srate",type=int,default=16384,help="Sampling rate")
parser.add_argument("--seglen", type=float,default=16., help="Default window size for processing.")
parser.add_argument("--start", type=int,default=None)
parser.add_argument("--stop", type=int,default=None)
parser.add_argument("--fref", dest='fref', type=float, default=0.0, help="Waveform reference frequency [template]. Required, default is 0 (coalescence).")
parser.add_argument("--incl",default=None,help="Set the inclination of L (at fref). Particularly helpful for aligned spin tests")
parser.add_argument("--mass1",default=10,type=float,help='Mass 1 (solar masses)')
parser.add_argument("--mass2",default=1.4,type=float,help='Mass 2 (solar masses)')
parser.add_argument("--verbose", action="store_true",default=False)
opts=  parser.parse_args()


# Generate signal
P = lalsimutils.ChooseWaveformParams()
P.deltaT = 1./opts.srate
P.radec = True  # use a real source with a real instrument
if not opts.inj:
    P.randomize(aligned_spin_Q=True,default_inclination=opts.incl)
    P.m1 = opts.mass1*lalsimutils.lsu_MSUN
    P.m2 = opts.mass2*lalsimutils.lsu_MSUN
    P.taper = lalsimutils.lsu_TAPER_START
    P.tref =1000000000  # default
    if opts.approx:
        P.approx = lalsim.GetApproximantFromString(str(opts.approx))
    else:
        P.approx = lalsim.GetApproximantFromString("SpinTaylorT2")
else:
    from ligo.lw import lsctables, table, utils # check all are needed

    filename = opts.inj
    event = opts.event_id
    xmldoc = utils.load_filename(filename, verbose = True, contenthandler =lalsimutils.cthdler)
    sim_inspiral_table = table.get_table(xmldoc, lsctables.SimInspiralTable.tableName)
    P.copy_sim_inspiral(sim_inspiral_table[int(event)])
    P.taper = lalsimutils.lsu_TAPER_START
    if opts.approx:
        P.approx = lalsim.GetApproximantFromString(str(opts.approx))
P.taper = lalsimutils.lsu_TAPER_START  # force taper
P.detector = opts.instrument
if opts.approx == "EccentricTD":
    P.phaseO = 3
P.print_params()


T_est = lalsimutils.estimateWaveformDuration(P)
T_est = P.deltaT*lalsimutils.nextPow2(T_est/P.deltaT)
if T_est < opts.seglen:
    T_est =opts.seglen
P.deltaF = 1./T_est
print(" Duration ", T_est)
if T_est < opts.seglen:
    print(" Buffer length too short, automating retuning forced ")


# Generate signal
hoft = lalsimutils.hoft(P)   # include translation of source, but NOT interpolation onto regular time grid
# zero pad to be opts.seglen long, if necessary
if opts.seglen/hoft.deltaT > hoft.data.length:
    TDlenGoal = int(opts.seglen/hoft.deltaT)
    hoft = lal.ResizeREAL8TimeSeries(hoft, 0, TDlenGoal)

# zero pad some more on either side, to make sure the segment covers start to stop
if opts.start and hoft.epoch > opts.start:
    nToAddBefore = int((float(hoft.epoch)-opts.start)/hoft.deltaT)
    # hoft.epoch - nToAddBefore*hoft.deltaT  # this is close to the epoch, but not quite ... we are adjusting it to be within 1 time sample
    print(nToAddBefore, hoft.data.length)
    ht = lal.CreateREAL8TimeSeries("Template h(t)", 
            opts.start , 0, hoft.deltaT, lalsimutils.lsu_DimensionlessUnit, 
            hoft.data.length+nToAddBefore)
    ht.data.data = np.zeros(ht.data.length)  # clear
    ht.data.data[nToAddBefore:nToAddBefore+hoft.data.length] = hoft.data.data
    hoft = ht

if opts.stop and hoft.epoch+hoft.data.length*hoft.deltaT < opts.stop:
    nToAddAtEnd = int( (-(hoft.epoch+hoft.data.length*hoft.deltaT)+opts.stop)/hoft.deltaT)
    print("Padding end ", nToAddAtEnd, hoft.data.length)
    hoft = lal.ResizeREAL8TimeSeries(hoft,0, int(hoft.data.length+nToAddAtEnd))

# Import background data, if needed, and add it
if not ( opts.background_cache is None  or opts.background_channel is None and opts.start is None and opts.stop is None):
    # Don't use specific times
    start = opts.start
    stop = opts.stop
    hoft_bg = lalsimutils.frame_data_to_hoft(opts.background_cache,opts.background_channel,start,stop)
    hoft.data.data += hoft_bg.data.data  # add noise
# elif opts.background_add_from_gaussian_psd:
#     # Retrieve PSDs
#     psd_here=None
#     for inst, psdf in map(lambda c: c.split("="), opts.psd_file):
#         if inst == opts.instrument:
#             print "Reading PSD for instrument %s from %s" % (inst, psdf)
#             psd_here = lalsimutils.get_psd_series_from_xmldoc(psdf, inst)

#     # Generate white noise
#     ht_noise = lal.CreateREAL8TimeSeries("Template h(t)", 
#             hoft.epoch, 0, hoft.deltaT, lalsimutils.lsu_DimensionlessUnit, 
#             hoft.data.length)
#     ht_noise.data.data = np.random.normal(0,1,size=ht_noise.data.length)

#     # Fourier transform
#     hF_noise = lalsimutils.DataFourierREAL8(ht_noise)


channel = opts.instrument+":FAKE-STRAIN"

tstart = int(hoft.epoch)
duration = int(round(hoft.data.length*hoft.deltaT))
if not opts.fname:
    fname = opts.instrument.replace("1","")+"-fake_strain-"+str(tstart)+"-"+str(duration)+".gwf"

print("Writing signal with ", hoft.data.length*hoft.deltaT, " to file ", fname)
lalsimutils.hoft_to_frame_data(fname,channel,hoft)

# TEST: Confirm it works by reading the frame
if opts.verbose:
    print(" -----  Plotting data ------ ")
    import os
    from matplotlib import pyplot as plt
    # First must create corresponding cache file
    os.system("echo "+ fname+ " | lalapps_path2cache   > test.cache")
    # Now I can read it
    # Beware that the results are OFFSET FROM ONE ANOTHER due to PADDING,
    #    but that the time associations are correct
    hoft2 = lalsimutils.frame_data_to_hoft("test.cache", channel)
    tvals2 = (float(hoft2.epoch) - float(P.tref)) +  np.arange(hoft2.data.length)*hoft2.deltaT
    tvals = (float(hoft.epoch) - float(P.tref)) +  np.arange(hoft.data.length)*hoft.deltaT
    plt.plot(tvals2,hoft2.data.data,label='Fr')
    plt.plot(tvals,hoft.data.data,label='orig')
    plt.legend(); #plt.show()
    plt.savefig("injected-data_"+opts.instrument +".png")




import numpy as np


def extract_combination_from_LI(samples_LI, p):
    """
    extract_combination_from_LI
      - reads in known columns from posterior samples
      - for selected known combinations not always available, it will compute them from standard quantities
    """
    if p in samples_LI.dtype.names:  # e.g., we have precomputed it
        return samples_LI[p]
    if p in remap_ILE_2_LI.keys():
       if remap_ILE_2_LI[p] in samples_LI.dtype.names:
         return samples_LI[ remap_ILE_2_LI[p] ]
    # Return cartesian components of spin1, spin2.  NOTE: I may already populate these quantities in 'Add important quantities'
    if p == 'chiz_plus':
        print(" Transforming ")
        if 'a1z' in samples_LI.dtype.names:
            return (samples_LI['a1z']+ samples_LI['a2z'])/2.
        if 'theta1' in samples_LI.dtype.names:
            return (samples_LI['a1']*np.cos(samples_LI['theta1']) + samples_LI['a2']*np.cos(samples_LI['theta2']) )/2.
#        return (samples_LI['a1']+ samples_LI['a2'])/2.
    if p == 'chiz_minus':
        print(" Transforming ")
        if 'a1z' in samples_LI.dtype.names:
            return (samples_LI['a1z']- samples_LI['a2z'])/2.
        if 'theta1' in samples_LI.dtype.names:
            return (samples_LI['a1']*np.cos(samples_LI['theta1']) - samples_LI['a2']*np.cos(samples_LI['theta2']) )/2.
#        return (samples_LI['a1']- samples_LI['a2'])/2.
    if  'theta1' in samples_LI.dtype.names:
        if p == 's1x':
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) * np.cos( samples_LI['phi1'])
        if p == 's1y' :
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) * np.sin( samples_LI['phi1'])
        if p == 's2x':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) * np.cos( samples_LI['phi2'])
        if p == 's2y':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) * np.sin( samples_LI['phi2'])
        if p == 'chi1_perp' :
            return samples_LI["a1"]*np.sin(samples_LI[ 'theta1']) 
        if p == 'chi2_perp':
            return samples_LI["a2"]*np.sin(samples_LI[ 'theta2']) 
    if 'lambdat' in samples_LI.dtype.names:  # LI does sampling in these tidal coordinates
        lambda1, lambda2 = lalsimutils.tidal_lambda_from_tilde(samples_LI["m1"], samples_LI["m2"], samples_LI["lambdat"], samples_LI["dlambdat"])
        if p == "lambda1":
            return lambda1
        if p == "lambda2":
            return lambda2
    if p == 'delta' or p=='delta_mc':
        return (samples_LI['m1']  - samples_LI['m2'])/((samples_LI['m1']  + samples_LI['m2']))
    # Return cartesian components of Lhat
    if p == 'product(sin_beta,sin_phiJL)':
        return np.sin(samples_LI[ remap_ILE_2_LI['beta'] ]) * np.sin(  samples_LI['phi_jl'])
    if p == 'product(sin_beta,cos_phiJL)':
        return np.sin(samples_LI[ remap_ILE_2_LI['beta'] ]) * np.cos(  samples_LI['phi_jl'])

    print(" No access for parameter ", p)
    return np.zeros(len(samples_LI['m1']))  # to avoid causing a hard failure

def add_field(a, descr):
    """Return a new array that is like "a", but has additional fields.

    Arguments:
      a     -- a structured numpy array
      descr -- a numpy type description of the new fields

    The contents of "a" are copied over to the appropriate fields in
    the new array, whereas the new fields are uninitialized.  The
    arguments are not modified.

    >>> sa = numpy.array([(1, 'Foo'), (2, 'Bar')], \
                         dtype=[('id', int), ('name', 'S3')])
    >>> sa.dtype.descr == numpy.dtype([('id', int), ('name', 'S3')])
    True
    >>> sb = add_field(sa, [('score', float)])
    >>> sb.dtype.descr == numpy.dtype([('id', int), ('name', 'S3'), \
                                       ('score', float)])
    True
    >>> numpy.all(sa['id'] == sb['id'])
    True
    >>> numpy.all(sa['name'] == sb['name'])
    True
    """
    if a.dtype.fields is None:
        raise ValueError("`A' must be a structured numpy array")
    b = np.empty(a.shape, dtype=a.dtype.descr + descr)
    for name in a.dtype.names:
        b[name] = a[name]
    return b






# gp.py: Very primitive gaussian process regression module.  Returns an object
#
#  - Prediction:   K(x,x').inv(Kmtx).y.  We store and precompute inv(Kmtx).y

import numpy as np  
from sklearn import gaussian_process

def squared_exponential_kernel(dx,h,sigma0,sigmab):
    ret  = np.power(sigma0,2) * np.exp(- np.power(dx,2)/(2.* np.power(h,2)))
    ret += np.where(np.abs(dx) < 1e-9, np.ones(dx.shape)*np.power(sigmab,2), np.zeros(dx.shape) )
    return ret


class GaussianProcess1d:
    """
    Very simple gaussian process model. Does not take into account errors that vary with position.
    """
    def __init__(self, dat1d,sigma0,sigmab=1e-5,h=None):
        self.h = h
        self.sigma0 = sigma0
        self.sigmab = sigmab
        x= self.x = dat1d[:,0]
        self.y = dat1d[:,1]
        if self.h == None:
            hmin = np.min(x - np.roll(x,1)) # minimum spacing, roughly
            self.h = np.max([hmin, 3*(np.max(x) - np.min(x))/len(x)])
        x = dat1d[:,0]
        xv,xpv = np.meshgrid(x,x)
        xMinusXp = xv - xpv
        Kmtx = squared_exponential_kernel(xMinusXp,self.h,self.sigma0,self.sigmab)
#        print "Det", np.linalg.det(Kmtx)
        KmtxInv = np.linalg.inv(Kmtx)
        KmtxInvY = np.matrix(KmtxInv)*np.matrix( np.reshape(dat1d[:,1], (len(x),1)))
        self.KmtxInvY = np.matrix(np.reshape(KmtxInvY, (len(x),1)))
    def predict(self, x_new):
        xv,xpv = np.meshgrid(x_new,self.x)
        xMinusXp = xv - xpv
        Kmtx_new = np.matrix(squared_exponential_kernel(xMinusXp,self.h,self.sigma0,self.sigmab).T)
#        print Kmtx_new.shape, self.KmtxInvY.shape
        return Kmtx_new* self.KmtxInvY



#! /usr/bin/env python
# util_WriteInjectionFile.py
# 
#  Lets user create injection xml files from the command line.  More efficient than monoblock python codes
#  User has option to create parameters identical to NR injections (i.e., correct spin values)
#
# EXAMPLES
#   /util_WriteInjectionFile.py  --group Sequence-SXS-All --param 1 --parameter mtot --parameter-value 70 --parameter dist --parameter-value 1000



import argparse
import sys
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lal
import functools

parser = argparse.ArgumentParser()
# Parameters
parser.add_argument("--parameter", action='append', help='Explicit list of parameters to use')
parser.add_argument("--parameter-value", action="append",help="Explicit list of parameter values to use")
parser.add_argument("--fname",default="mdc",help="Injection xml name (.xml.gz will be added)")
parser.add_argument("--approximant",default="TaylorT4",help="Set approximant. Useful for creating LAL comparison injections.")
# Add option to use NR waveforms!
parser.add_argument("--group",default=None)
parser.add_argument("--param",default=None)
opts=  parser.parse_args()

hasNR = False
if opts.group:
    import NRWaveformCatalogManager3 as nrwf
    hasNR=True


P=None
if not hasNR:
    P=lalsimutils.ChooseWaveformParams()
else:
    param = opts.param
    if nrwf.internal_ParametersAreExpressions[opts.group]:
        param = eval(param)
    wfP = nrwf.WaveformModeCatalog(opts.group, param, metadata_only=True)
    P = wfP.P

# default sky location, etc [will be overridden]
P.radec = True
P.theta = np.pi/2
P.phi = 0
P.phiref = 0
P.psi = 0
P.approx = lalsimutils.lalsim.GetApproximantFromString(opts.approximant)  # allow user to override the approx setting. Important for NR followup, where no approx set in sim_xml!

param_names = opts.parameter
for param in param_names:
    # Check if in the valid list
    if not(param in lalsimutils.valid_params):
        print(' Invalid param ', param, ' not in ', lalsimutils.valid_params)
        sys.exit(0)

param_values = []
if len(param_names) == len(opts.parameter_value):
    param_values = list(map(eval, opts.parameter_value))
for p in ['mc', 'm1', 'm2', 'mtot']:
    if p in param_names:
        indx = param_names.index(p)
        param_values[indx]= np.array(param_values[indx])* lal.MSUN_SI

for p in ['dist']:
    if p in param_names:
        indx = param_names.index(p)
        param_values[indx]= np.array(param_values[indx])* lal.PC_SI*1e6


for indx in np.arange(len(param_values)):
    P.assign_param(param_names[indx], param_values[indx])

print(" ------- WRITING INJECTION FILE ----- ")
P.print_params()

P_list = [P]
lalsimutils.ChooseWaveformParams_array_to_xml(P_list, fname=opts.fname, fref=P.fref)




import sys
import math
import bisect
from collections import defaultdict

import numpy as np

import itertools
import functools

#from statutils import cumvar

from multiprocessing import Pool

from . import MonteCarloEnsemble as monte_carlo

__author__ = "Ben Champion"

rosDebugMessages = True

class NanOrInf(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class MCSampler(object):
    """
    Class to define a set of parameter names, limits, and probability densities.
    """

    @staticmethod
    def match_params_from_args(args, params):
        """
        Given two unordered sets of parameters, one a set of all "basic" elements
        (strings) possible, and one a set of elements both "basic" strings and
        "combined" (basic strings in tuples), determine whether the sets are equivalent
        if no basic element is repeated.

        e.g. set A ?= set B

        ("a", "b", "c") ?= ("a", "b", "c") ==> True
        (("a", "b", "c")) ?= ("a", "b", "c") ==> True
        (("a", "b"), "d")) ?= ("a", "b", "c") ==> False  # basic element 'd' not in set B
        (("a", "b"), "d")) ?= ("a", "b", "d", "c") ==> False  # not all elements in set B 
        represented in set A
        """
        not_common = set(args) ^ set(params)
        if len(not_common) == 0:
            # All params match
            return True
        if all([not isinstance(i, tuple) for i in not_common]):
            # The only way this is possible is if there are
            # no extraneous params in args
            return False

        to_match = [i for i in not_common if not isinstance(i, tuple)]
        against = [i for i in not_common if isinstance(i, tuple)]
        
        matched = []
        import itertools
        for i in range(2, max(list(map(len, against)))+1):
            matched.extend([t for t in itertools.permutations(to_match, i) if t in against])
        return (set(matched) ^ set(against)) == set()


    def __init__(self):
        # Total number of samples drawn
        self.ntotal = 0
        # Samples per iteration
        self.n = 0
        # Parameter names
        self.params = set()
        self.params_ordered = []  # keep them in order. Important to break likelihood function need for names
        # parameter -> pdf function object
        self.pdf = {}
        # If the pdfs aren't normalized, this will hold the normalization 
        # constant
        self._pdf_norm = defaultdict(lambda: 1)
        # Cache for the sampling points
        self._rvs = {}
        # parameter -> cdf^{-1} function object
        self.cdf = {}
        self.cdf_inv = {}
        # params for left and right limits
        self.llim, self.rlim = {}, {}
        # Keep track of the adaptive parameters
        self.adaptive = []

        # Keep track of the adaptive parameter 1-D marginalizations
        self._hist = {}

        # MEASURES (=priors): ROS needs these at the sampler level, to clearly separate their effects
        # ASSUMES the user insures they are normalized
        self.prior_pdf = {}

        self.func = None
        self.sample_format = None
        self.curr_args = None


    def clear(self):
        """
        Clear out the parameters and their settings, as well as clear the sample cache.
        """
        self.params = set()
        self.params_ordered = []
        self.pdf = {}
        self._pdf_norm = defaultdict(lambda: 1.0)
        self._rvs = {}
        self._hist = {}
        self.cdf = {}
        self.cdf_inv = {}
        self.llim = {}
        self.rlim = {}
        self.adaptive = []

    def add_parameter(self, params, pdf=None,  cdf_inv=None, left_limit=None, right_limit=None, 
                        prior_pdf=None, adaptive_sampling=False):
        """
        Add one (or more) parameters to sample dimensions. params is either a string 
        describing the parameter, or a tuple of strings. The tuple will indicate to 
        the sampler that these parameters must be sampled together. left_limit and 
        right_limit are on the infinite interval by default, but can and probably should 
        be specified. If several params are given, left_limit, and right_limit must be a 
        set of tuples with corresponding length. Sampling PDF is required, and if not 
        provided, the cdf inverse function will be determined numerically from the 
        sampling PDF.
        """
        self.params.add(params) # does NOT preserve order in which parameters are provided
        self.params_ordered.append(params)
        if rosDebugMessages:
            print(" mcsampler: Adding parameter ", params, " with limits ", [left_limit, right_limit])
        if isinstance(params, tuple):
            assert all([lim[0] < lim[1] for lim in zip(left_limit, right_limit)])
            if left_limit is None:
                self.llim[params] = list(float("-inf"))*len(params)
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = list(float("+inf"))*len(params)
            else:
                self.rlim[params] = right_limit
            if prior_pdf is not None:
                for p in params:
                    self.prior_pdf[p] = prior_pdf
        else:
            assert left_limit < right_limit
            if left_limit is None:
                self.llim[params] = float("-inf")
            else:
                self.llim[params] = left_limit
            if right_limit is None:
                self.rlim[params] = float("+inf")
            else:
                self.rlim[params] = right_limit
            if prior_pdf is not None:
                self.prior_pdf[params] = prior_pdf

    def evaluate(self, samples):
        '''
        Interfaces between monte_carlo_integrator sample format (1 (n x d) array)
        and likelihood function sample format (d 1D arrays in a list)
        '''
        # integrand expects a list of 1D rows
        temp = []
        for index in range(len(self.curr_args)):
            temp.append(samples[:,index])
        temp_ret = self.func(*temp)
        return np.rot90([temp_ret], -1) # monte_carlo_integrator expects a column


    def calc_pdf(self, samples):
        '''
        Similar to evaluate(), interfaces between sample formats. Must also handle
        possibility of no prior for one of more dimensions
        '''
        n, _ = samples.shape
        temp_ret = np.ones((n, 1))
        # pdf functions expect 1D rows
        for index in range(len(self.curr_args)):
            if self.curr_args[index] in self.prior_pdf:
                pdf_func = self.prior_pdf[self.curr_args[index]]
                temp_samples = samples[:,index]
                # monte carlo integrator expects a column
                temp_ret *= np.rot90([pdf_func(temp_samples)], -1)
        return temp_ret


    def integrate(self, func, *args,**kwargs):
        '''
        Integrate the specified function over the specified parameters.

        func: function to integrate

        args: list of parameters to integrate over

        direct_eval (bool): whether func can be evaluated directly with monte_carlo_integrator
        format or not

        n_comp: number of gaussian components for model

        n: number of samples per iteration

        nmax: maximum number of samples for all iterations

        write_to_file (bool): write data to file

        gmm_dict: dictionary of dimensions and mixture models (see monte_carlo_integrator
        documentation for more)

        var_thresh: result variance threshold for termination

        min_iter: minimum number of integrator iterations

        max_iter: maximum number of integrator iterations

        neff: eff_samp cutoff for termination

        reflect (bool): whether or not to reflect samples over boundaries (you should
        basically never use this, it's really slow)

        mcsamp_func: function to be executed before mcsampler_new terminates (for example,
        to print results or debugging info)

        integrator_func: function to be executed each iteration of the integrator (for
        example, to print intermediate results)

        proc_count: size of multiprocessing pool. set to None to not use multiprocessing
        tempering_exp -- Exponent to raise the weights of the 1-D marginalized histograms for adaptive sampling prior generation, by default it is 0 which will turn off adaptive sampling regardless of other settings

        max_err : Maximum number of errors allowed for GMM sampler
        '''
        nmax = kwargs["nmax"] if "nmax" in kwargs else 1e6
        neff = kwargs["neff"] if "neff" in kwargs else 1000
        n = kwargs["n"] if "n" in kwargs else min(1000, nmax)  # chunk size
        n_comp = kwargs["n_comp"] if "n_comp" in kwargs else 1
        if 'gmm_dict' in list(kwargs.keys()):
            gmm_dict = kwargs['gmm_dict']  # required
        else:
            gmm_dict = None
        reflect = kwargs['reflect'] if "reflect" in kwargs else False
        integrator_func  = kwargs['integrator_func'] if "integrator_func" in kwargs  else None
        mcsamp_func  = kwargs['mcsamp_func'] if "mcsamp_func" in kwargs  else None
        proc_count = kwargs['proc_count'] if "proc_count" in kwargs else None
        direct_eval = kwargs['direct_eval'] if "direct_eval" in kwargs else False
        min_iter = kwargs['min_iter'] if "min_iter" in kwargs else 10
        max_iter = kwargs['max_iter'] if "max_iter" in kwargs else 20
        var_thresh = kwargs['var_thres'] if "var_thresh" in kwargs else 0.05
        write_to_file = kwargs['write_to_file'] if "write_to_file" in kwargs else False
        correlate_all_dims = kwargs['correlate_all_dims'] if  "correlate_all_dims" in kwargs else False
        gmm_adapt = kwargs['gmm_adapt'] if "gmm_adapt" in kwargs else None
        gmm_epsilon = kwargs['gmm_epsilon'] if "gmm_epsilon" in kwargs else None
        L_cutoff = kwargs["L_cutoff"] if "L_cutoff" in kwargs else None
        tempering_exp = kwargs["tempering_exp"] if "tempering_exp" in kwargs else 0.0

        max_err = kwargs["max_err"] if "max_err" in kwargs else 10  # default

        
        # set up a lot of preliminary stuff
        self.func = func
        self.curr_args = args
        if n_comp is None:
            print('No n_comp given, assuming 1 component per dimension')
            n_comp = 1
        dim = len(args)
        bounds=[]
        for param in args:
            bounds.append([self.llim[param], self.rlim[param]])
        raw_bounds = np.array(bounds)

        bounds=None
        # generate default gmm_dict if not specified
        if gmm_dict is None:
            bounds = {}
            for indx in np.arange(len(raw_bounds)):
                bounds[(indx,)] = raw_bounds[indx]
            bounds=raw_bounds
            if correlate_all_dims:
                gmm_dict = {tuple(range(dim)):None}
                bounds = {tuple(np.arange(len(bounds))): raw_bounds}
            else:
                gmm_dict = {}
                for i in range(dim):
                    gmm_dict[(i,)] = None
        else:
            # create bounds that depend on the dimension specifiers in the gmm integrator
            bounds ={}
            for dims in gmm_dict:
                n_dims = len(dims)
                bounds_here = np.empty((n_dims,2))
                for indx in np.arange(n_dims):
                    bounds_here[indx] = raw_bounds[dims[indx]]  # pull out bounds index
                bounds[dims]=bounds_here
#            bounds = np.array(bounds)
        # do the integral

        integrator = monte_carlo.integrator(dim, bounds, gmm_dict, n_comp, n=n, prior=self.calc_pdf,
                         user_func=integrator_func, proc_count=proc_count,L_cutoff=L_cutoff,gmm_epsilon=gmm_epsilon,tempering_exp=tempering_exp) # reflect=reflect,
        if not direct_eval:
            func = self.evaluate
        integrator.integrate(func, min_iter=min_iter, max_iter=max_iter, var_thresh=var_thresh, neff=neff, nmax=nmax,max_err=max_err)

        # get results

        self.n = int(integrator.n)
        self.ntotal = int(integrator.ntotal)
        integral = integrator.integral
        error_squared = integrator.scaled_error_squared * np.exp(integrator.log_error_scale_factor)
        eff_samp = integrator.eff_samp
        sample_array = integrator.cumulative_samples
        value_array = np.exp(integrator.cumulative_values)  # stored as ln(integrand) !
        p_array = integrator.cumulative_p_s
        prior_array = integrator.cumulative_p

        # user-defined function
        if mcsamp_func is not None:
            mcsamp_func(self, integrator)

        # populate dictionary

        index = 0
        for param in args:
            self._rvs[param] = sample_array[:,index]
            index += 1
        self._rvs['joint_prior'] = prior_array
        self._rvs['joint_s_prior'] = p_array
        self._rvs['integrand'] = value_array

        # write data to file

        if write_to_file:
            dat_out = np.c_[sample_array, value_array, p_array]
            np.savetxt('mcsampler_data.txt', dat_out,
                        header=" ".join(['sample_array', 'value_array', 'p_array']))

        return integral, error_squared, eff_samp, {}


def inv_uniform_cdf(a, b, x):
    return (b-a)*x+a

def gauss_samp(mu, std, x):
    return 1.0/np.sqrt(2*np.pi*std**2)*np.exp(-(x-mu)**2/2/std**2)

def gauss_samp_withfloor(mu, std, myfloor, x):
    return 1.0/np.sqrt(2*np.pi*std**2)*np.exp(-(x-mu)**2/2/std**2) + myfloor

#gauss_samp_withfloor_vector = np.vectorize(gauss_samp_withfloor,excluded=['mu','std','myfloor'],otypes=[np.float])
gauss_samp_withfloor_vector = np.vectorize(gauss_samp_withfloor,otypes=[np.float])


# Mass ratio. PDF propto 1/(1+q)^2.  Defined so mass ratio is < 1
# expr = Integrate[1/(1 + q)^2, q]
# scale = (expr /. q -> qmax )  - (expr /. q -> qmin)
# (expr - (expr /. q -> qmin))/scale == x // Simplify
# q /. Solve[%, q][[1]] // Simplify
# % // CForm
def q_samp_vector(qmin,qmax,x):
    scale = 1./(1+qmin) - 1./(1+qmax)
    return 1/np.power((1+x),2)/scale
def q_cdf_inv_vector(qmin,qmax,x):
    return np.array((qmin + qmax*qmin + qmax*x - qmin*x)/(1 + qmax - qmax*x + qmin*x),dtype=np.float128)

# total mass. Assumed used with q.  2M/Mmax^2-Mmin^2
def M_samp_vector(Mmin,Mmax,x):
    scale = 2./(Mmax**2 - Mmin**2)
    return x*scale


def cos_samp(x):
        return np.sin(x)/2   # x from 0, pi

def dec_samp(x):
        return np.sin(x+np.pi/2)/2   # x from 0, pi

cos_samp_vector = np.vectorize(cos_samp,otypes=[np.float])
dec_samp_vector = np.vectorize(dec_samp,otypes=[np.float])
def cos_samp_cdf_inv_vector(p):
    return np.arccos( 2*p-1)   # returns from 0 to pi
def dec_samp_cdf_inv_vector(p):
    return np.arccos(2*p-1) - np.pi/2  # target from -pi/2 to pi/2


def pseudo_dist_samp(r0,r):
        return r*r*np.exp( - (r0/r)*(r0/r)/2. + r0/r)+0.01  # put a floor on probability, so we converge. Note this floor only cuts out NEARBY distances

#pseudo_dist_samp_vector = np.vectorize(pseudo_dist_samp,excluded=['r0'],otypes=[np.float])
pseudo_dist_samp_vector = np.vectorize(pseudo_dist_samp,otypes=[np.float])

def delta_func_pdf(x_0, x):
    return 1.0 if x == x_0 else 0.0

delta_func_pdf_vector = np.vectorize(delta_func_pdf, otypes=[np.float])

def delta_func_samp(x_0, x):
    return x_0

delta_func_samp_vector = np.vectorize(delta_func_samp, otypes=[np.float])

class HealPixSampler(object):
    """
    Class to sample the sky using a FITS healpix map. Equivalent to a joint 2-D pdf in RA and dec.
    """

    @staticmethod
    def thph2decra(th, ph):
        """
        theta/phi to RA/dec
        theta (north to south) (0, pi)
        phi (east to west) (0, 2*pi)
        declination: north pole = pi/2, south pole = -pi/2
        right ascension: (0, 2*pi)
        
        dec = pi/2 - theta
        ra = phi
        """
        return np.pi/2-th, ph

    @staticmethod
    def decra2thph(dec, ra):
        """
        theta/phi to RA/dec
        theta (north to south) (0, pi)
        phi (east to west) (0, 2*pi)
        declination: north pole = pi/2, south pole = -pi/2
        right ascension: (0, 2*pi)
        
        theta = pi/2 - dec
        ra = phi
        """
        return np.pi/2-dec, ra

    def __init__(self, skymap, massp=1.0):
        self.skymap = skymap
        self._massp = massp
        self.renormalize()

    @property
    def massp(self):
        return self._massp

    @massp.setter
    def massp(self, value):
        assert 0 <= value <= 1
        self._massp = value
        norm = self.renormalize()

    def renormalize(self):
        """
        Identify the points contributing to the overall cumulative probability distribution, and set the proper normalization.
        """
        res = healpy.npix2nside(len(self.skymap))
        self.pdf_sorted = sorted([(p, i) for i, p in enumerate(self.skymap)], reverse=True)
        self.valid_points_decra = []
        cdf, np = 0, 0
        for p, i in self.pdf_sorted:
            if p == 0:
                continue # Can't have a zero prior
            self.valid_points_decra.append(HealPixSampler.thph2decra(*healpy.pix2ang(res, i)))
            cdf += p
            if cdf > self._massp:
                break
        self._renorm = cdf
        # reset to indicate we'd need to recalculate this
        self.valid_points_hist = None
        return self._renorm

    def __expand_valid(self, min_p=1e-7):
        #
        # Determine what the 'quanta' of probabilty is
        #
        if self._massp == 1.0:
            # This is to ensure we don't blow away everything because the map
            # is very spread out
            min_p = min(min_p, max(self.skymap))
        else:
            # NOTE: Only valid if CDF descending order is kept
            min_p = self.pseudo_pdf(*self.valid_points_decra[-1])

        self.valid_points_hist = []
        ns = healpy.npix2nside(len(self.skymap))

        # Renormalize first so that the vector histogram is properly normalized
        self._renorm = 0
        # Account for probability lost due to cut off
        for i, v in enumerate(self.skymap >= min_p):
            self._renorm += self.skymap[i] if v else 0

        for pt in self.valid_points_decra:
            th, ph = HealPixSampler.decra2thph(pt[0], pt[1])
            pix = healpy.ang2pix(ns, th, ph)
            if self.skymap[pix] < min_p:
                continue
            self.valid_points_hist.extend([pt]*int(round(self.pseudo_pdf(*pt)/min_p)))
        self.valid_points_hist = np.array(self.valid_points_hist).T

    def pseudo_pdf(self, dec_in, ra_in):
        """
        Return pixel probability for a given dec_in and ra_in. Note, uses healpy functions to identify correct pixel.
        """
        th, ph = HealPixSampler.decra2thph(dec_in, ra_in)
        res = healpy.npix2nside(len(self.skymap))
        return self.skymap[healpy.ang2pix(res, th, ph)]/self._renorm

    def pseudo_cdf_inverse(self, dec_in=None, ra_in=None, ndraws=1, stype='vecthist'):
        """
        Select points from the skymap with a distribution following its corresponding pixel probability. If dec_in, ra_in are suupplied, they are ignored except that their shape is reproduced. If ndraws is supplied, that will set the shape. Will return a 2xN np array of the (dec, ra) values.
        stype controls the type of sampling done to retrieve points. Valid choices are
        'rejsamp': Rejection sampling: accurate but slow
        'vecthist': Expands a set of points into a larger vector with the multiplicity of the points in the vector corresponding roughly to the probability of drawing that point. Because this is not an exact representation of the proability, some points may not be represented at all (less than quantum of minimum probability) or inaccurately (a significant fraction of the fundamental quantum).
        """

        if ra_in is not None:
            ndraws = len(ra_in)
        if ra_in is None:
            ra_in, dec_in = np.zeros((2, ndraws))

        if stype == 'rejsamp':
            # FIXME: This is only valid under descending ordered CDF summation
            ceiling = max(self.skymap)
            i, np = 0, len(self.valid_points_decra)
            while i < len(ra_in):
                rnd_n = np.random.randint(0, np)
                trial = np.random.uniform(0, ceiling)
                if trial <= self.pseudo_pdf(*self.valid_points_decra[rnd_n]):
                    dec_in[i], ra_in[i] = self.valid_points_decra[rnd_n]
                    i += 1
            return np.array([dec_in, ra_in])
        elif stype == 'vecthist':
            if self.valid_points_hist is None:
                self.__expand_valid()
            np = self.valid_points_hist.shape[1]
            rnd_n = np.random.randint(0, np, len(ra_in))
            dec_in, ra_in = self.valid_points_hist[:,rnd_n]
            return np.array([dec_in, ra_in])
        else:
            raise ValueError("%s is not a recgonized sampling type" % stype)

#pseudo_dist_samp_vector = np.vectorize(pseudo_dist_samp,excluded=['r0'],otypes=[np.float])
pseudo_dist_samp_vector = np.vectorize(pseudo_dist_samp,otypes=[np.float])


def sanityCheckSamplerIntegrateUnity(sampler,*args,**kwargs):
        return sampler.integrate(lambda *args: 1,*args,**kwargs)

###
### CONVERGENCE TESTS
###


# neff by another name:
#    - value: tests for 'smooth' 1-d cumulative distributions
#    - require  require the most significant-weighted point be less than p of all cumulative probability
#    - this test is *equivalent* to neff > 1/p
#    - provided to illustrate the interface
def convergence_test_MostSignificantPoint(pcut, rvs, params):
    weights = rvs["weights"] #rvs["integrand"]* rvs["joint_prior"]/rvs["joint_s_prior"]
    indxmax = np.argmax(weights)
    wtSum = np.sum(weights)
    return  weights[indxmax]/wtSum < pcut


# normality test: is the MC integral normally distributed, with a small standard deviation?
#    - value: tests for converged integral
#    - arguments: 
#         - ncopies:               # of sub-integrals
#         - pcutNormalTest     Threshold p-value for normality test
#         - sigmaCutErrorThreshold   Threshold relative error in the integral
#    - implement normality test on **log(integral)** since the log should also be normally distributed if well converged
#           - this helps us handle large orders-of-magnitude differences
#           - compatible with a *relative* error threshold on integral
#           - only works for *positive-definite* integrands
#    - other python normality tests:  
#          scipy.stats.shapiro
#          scipy.stats.anderson
#  WARNING:
#    - this test assumes *unsorted* past history: the 'ncopies' segments are assumed independent.
import scipy.stats as stats
def convergence_test_NormalSubIntegrals(ncopies, pcutNormalTest, sigmaCutRelativeErrorThreshold, rvs, params):
    weights = rvs["integrand"]* rvs["joint_prior"]/rvs["joint_s_prior"]  # rvs["weights"] # rvs["weights"] is *sorted* (side effect?), breaking test. Recalculated weights are not.  Use explicitly calculated weights until sorting effect identified
#    weights = weights /np.sum(weights)    # Keep original normalization, so the integral values printed to stdout have meaning relative to the overall integral value.  No change in code logic : this factor scales out (from the log, below)
    igrandValues = np.zeros(ncopies)
    len_part = np.int(len(weights)/ncopies)  # deprecated: np.floor->np.int
    for indx in np.arange(ncopies):
        igrandValues[indx] = np.log(np.mean(weights[indx*len_part:(indx+1)*len_part]))  # change to mean rather than sum, so sub-integrals have meaning
    igrandValues= np.sort(igrandValues)#[2:]                            # Sort.  Useful in reports
    valTest = stats.normaltest(igrandValues)[1]                              # small value is implausible
    igrandSigma = (np.std(igrandValues))/np.sqrt(ncopies)   # variance in *overall* integral, estimated from variance of sub-integrals
    print(" Test values on distribution of log evidence:  (gaussianity p-value; standard deviation of ln evidence) ", valTest, igrandSigma)
    print(" Ln(evidence) sub-integral values, as used in tests  : ", igrandValues)
    return valTest> pcutNormalTest and igrandSigma < sigmaCutRelativeErrorThreshold   # Test on left returns a small value if implausible. Hence pcut ->0 becomes increasingly difficult (and requires statistical accidents). Test on right requires relative error in integral also to be small when pcut is small.   FIXME: Give these variables two different names
    



#! /bin/bash
# construct_GraceDB_pipeline_wrapper <GraceDBID> <approximant>
#     Creates a new directory with the GraceDB event ID
#     Retrieves data
#     Generates a dag (and a 'test command') for processing that event

export ILE=`which integrate_likelihood_extrinsic`
export CME=`which compute_marginalized_likelihood`
export EWD=`which util_EstimateWaveformDuration.py`
echo Using ILE ${ILE} 
echo Using CME ${CME} 
echo Using EWD ${EWD}
mkdir -p $1
cd $1

export gid=$1
if [ -n "$2" ]; then
  export approximant=$2
else
  export approximant=TaylorT4
fi
export ldf_server=ldr.ligo.caltech.edu:443
if [ -n "${LIGO_DATAFIND_SERVER}" ]; then
    ldf_server=${LIGO_DATAFIND_SERVER}
fi
echo " Constructing directory for ", $gid, " using approximant ", $approximant

gracedb download $gid coinc.xml 
if [ ! -e coinc.xml ]; then   # Stop if not in GraceDB
    echo " No event ! "
    exit 1
fi
echo " GDB event ", $gid, " mass pairs: ",`ligolw_print -t sngl_inspiral -c mass1 -c mass2 coinc.xml | head -1`
gracedb download $gid skymap.fits.gz 
gracedb download ${gid} psd.xml.gz 


export MASS1=`ligolw_print -t sngl_inspiral -c mass1 coinc.xml | head -1`
export MASS2=`ligolw_print -t sngl_inspiral -c mass2 coinc.xml  | head -1`
export BETA=`util_AdaptiveExponent.py --Niter 4000 --coinc coinc.xml | tail -1`
if [ ! -n "$MASS1" ];  then  
   echo "FAILURE:  mass not populated"
   exit 1
fi

export tEvent=`ligolw_print -t coinc_inspiral -c end_time coinc.xml`
export tDuration=`${EWD} --coinc coinc.xml | tail -n 1`
export tBefore=`echo   "${tDuration}*1.1 + 8+2 " | bc `    # Buffer for inverse spectrum truncation plus safety. Just used for data retrieval.
export tAfter=${tBefore}   # Buffer for inverse spectrum truncation plus safety. Just used for data retrieval.
export DATA_START=`echo $tEvent - $tBefore | bc`
export DATA_END=`echo $tEvent + $tAfter | bc`
echo " Event ", $tEvent
echo " DATA_START ", $DATA_START
echo " DATA_END ", $DATA_END
echo " Duration ", $tDuration

if [ ! -n "$tDuration" ];  then  
   echo "FAILURE:  duration not populated"
   exit 1
fi



if [ ! -n "${LIGO_DATAFIND_SERVER}" ]; then 
  echo " Assuming user on laptop; retrieving data files ";
  ligo_data_find \
		-u file \
		--gaps \
		-o H \
		-t H1_ER_C00_L1 \
		--server ${ldf_server}\
		-s $DATA_START \
		-e $DATA_END > H_remote.cache;
   ligo_data_find \
		-u file \
		--gaps \
		-o L \
		-t L1_ER_C00_L1 \
		--server ${ldf_server}\
		-s $DATA_START \
		-e $DATA_END >L_remote.cache;
   ligo_data_find \
		-u file \
		--gaps \
		-o V \
		-t V1Online \
		--server ${ldf_server}\
		-s $DATA_START \
		-e $DATA_END > V_remote.cache;
   mkdir -p gdb-data-${gid};
   while read line; do 
			gsiscp ldas-pcdev1.ligo.caltech.edu:${line:16} gdb-data-${gid}/; \
   done < H_remote.cache;
   while read line; do 
        		gsiscp ldas-pcdev1.ligo.caltech.edu:${line:16} gdb-data-${gid}/; \
   done < L_remote.cache;
   while read line; do 
        		gsiscp ldas-pcdev1.ligo.caltech.edu:${line:16} gdb-data-${gid}/; \
   done < V_remote.cache;

   find . -name '*.gwf' | lalapps_path2cache > local.cache;

else
  echo " Assuming user on cluster "
  ligo_data_find \
		-u file \
		--gaps \
		-o H \
		-t H1_ER_C00_L1 \
		--server ${ldf_server}\
		-s $DATA_START \
		-e $DATA_END > H_remote.cache;
   ligo_data_find \
		-u file \
		--gaps \
		-o L \
		-t L1_ER_C00_L1 \
		--server ${ldf_server}\
		-s $DATA_START \
		-e $DATA_END >L_remote.cache;
   ligo_data_find \
		-u file \
		--gaps \
		-o V \
		-t V1Online \
		--server ${ldf_server}\
		-s $DATA_START \
		-e $DATA_END > V_remote.cache;

   cat ?_remote.cache |  util_CacheFileConvert.py > local.cache;

fi

# CHANNEL NAMES
#    - by default, use the ER channel names -- that is what I will use for now
INJ_CHANNEL_NAME=FAKE-STRAIN
GDB_V_INJ_CHANNEL_NAME=FAKE_h_16384Hz_4R


## Make DAG

# Write a command that has the same options but just evaluates the likliehood (marginalized) at the trigger masses.  Helpful for debugging setup
#    - Comments left to indicate kind of arguments ROS wants on master
echo  ${ILE}  --cache-file local.cache  --coinc coinc.xml   --channel-name H1=${INJ_CHANNEL_NAME} --channel-name L1=${INJ_CHANNEL_NAME} --channel-name V1=${GDB_V_INJ_CHANNEL_NAME} --psd-file "H1=psd.xml.gz" --psd-file "L1=psd.xml.gz" --psd-file "V1=psd.xml.gz"   --mass1 ${MASS1} --mass2 ${MASS2} --reference-freq 0 --save-samples    --time-marginalization --n-max 200000 --n-eff 1000 --output-file ile-gracedb-${gid}.xml.gz   --save-P 0.0001 --fmax 2000 --adapt-weight-exponent ${BETA} --adapt-floor-level 0.1 --n-chunk 4000 --approximant $approximant --convergence-tests-on  > testme-command.sh  #  --adapt-parameter right_ascension --adapt-parameter declination --adapt-parameter distance       --approximant $approximant
# add postprocessing
echo make_triplot ile-gracedb-${gid}.xml.gz -o ${gid}-triplot.pdf >> testme-command.sh
echo plot_integral ile-gracedb-${gid}.xml.gz --output integral.pdf >> testme-command.sh
chmod a+x testme-command.sh



# Write the actual DAG
#   - large n-max chosen for prototyping purposes.  Hopefully we will hit the n-eff limit before reaching it.
${CME} --cache-file local.cache  --coinc coinc.xml  --channel-name H1=${INJ_CHANNEL_NAME} --channel-name L1=${INJ_CHANNEL_NAME} --channel-name V1=${GDB_V_INJ_CHANNEL_NAME} --psd-file "H1=psd.xml.gz" --psd-file "L1=psd.xml.gz" --psd-file "V1=psd.xml.gz"   --mass1 ${MASS1} --mass2 ${MASS2}  --save-samples  --time-marginalization --n-max 1000000 --n-eff 1000 --output-file CME-${gid}.xml.gz   --save-P 0.0001  --n-copies 2 --fmax 2000 --adapt-weight-exponent ${BETA} --adapt-floor-level 0.1 --n-chunk 4000 --approximant $approximant --convergence-tests-on # --adapt-parameter right_ascension --adapt-parameter declination --adapt-parameter distance   --fmax 2000

# Write commands to do some useful postprocessing.  Ideally part of the dag.
#   -  convert the result to a flat ascii grid in m1,m2, lnL
#   -  convert the individual XML outputs to a single compressed tabular ascii file
#   -  make some detailed 2d plots.  These *should* be done by the DAG.  (We need to add a variable number of bins)
echo 'for i in CME-*.xml.gz; do ligolw_print -t sngl_inspiral -c mass1 -c mass2 -c snr -c tau0  -d ' ' $i; done > massgrid.txt &' > postprocess-massgrid.sh
cat >> postprocess-massgrid.sh <<EOF
convert_output_format_ile2inference  CME-*.xml.gz > flatfile-points.dat
postprocess_1d_cumulative --save-sampler-file flatfile
gzip flatfile-points.dat
cat ILE_MASS*.cache > net-ile.cache 
ligolw_sqlite CME-*.xml.gz -d net-ile.sqlite
EOF
chmod a+x postprocess-massgrid.sh

echo "===== TO SUBMIT THE DAG ====== "
echo condor_submit_dag `pwd`/dagtest-gracedb-${gid}-marginalize.dag

echo "===== TO TEST THE UNDERLYING INFRASTRUCTURE USING THIS DATA ====== "
echo  `pwd`/testme-command.sh 
